#include "stock.h"
#include "Pkt.h"
#include "swapbytes.h"

static int FullPrecision = 0 ; 

int 
showPktFullPrecisionTime(int i)
{
    int prev ; 
    prev = FullPrecision ;
    FullPrecision = i ; 
    return prev ; 
}

#define IS_PRINT(C) (((C)>=32) && (((C)<127) || ((C)=='\n') || ((C)=='\t') || ((C)=='\r')))

static int
ok2print(char *s, int n) 
{
    int i, c ; 
    for (i=0 ; i<n ; i++ ) { 
	c = s[i] ; 
	if ( c < 0x20 ) { 
	    if ( c != '\n' && c != '\t' ) { 
		return 0 ; 
	    } 
	} else if ( c > 0x7e ) { 
	    return 0 ; 
	}
    }
    return 1 ; 
}

static void
show_stash ( FILE *file, Packet *pkt, int mode)
{
    char aline[81] ; 
    switch (mode) { 
      default :
      case PKT_PEEK:
	if (ok2print(pkt->string, sizeof(aline))) { 
	    strncpy(aline, pkt->string, sizeof(aline)-1) ; 
	    aline[sizeof(aline)-1] = 0 ; 
	    elog_fprintf(file, "%s\n", aline ) ;
	} else { 
	    hexdump (file, pkt->string, MIN(48, pkt->string_size)) ; 
	}
	break ;
      case PKT_NOSAMPLES:
	break ;
      case PKT_UNSTUFF:
        elog_fprintf ( file, "%s\n", okstr(pkt->string) ) ; 
	break;
    }
}

static void
show_fl (int pktid, char *srcname, double pkttime, char *packet, int nbytes, Packet *unstuffed, FILE *file, int mode ) 
{
    char *id, *cp ;
    int i, seq ; 
    id = strdup (unstuffed->parts.src_subcode) ; 
    cp = strchr(id, '/') ; 
    if ( cp != 0 ) { 
	*cp = 0 ; 
    }
    elog_fprintf(file, "    id       = %s\n", id ) ; 
    free(id) ;
    seq = unstuffed->string_size ;
    elog_fprintf(file, "    sequence = %8d\n", seq ) ; 
    if ( seq == 0 ) { 
	elog_fprintf(file, "    filename   = %s\n", okstr(unstuffed->dfile) ) ; 
	elog_fprintf(file, "    mode       = %8o\n", unstuffed->nchannels ) ;
    }
    elog_fprintf(file, "    digest   = " ) ; 
    for (i = 0; i < 16; i++) {
        printf ("%02x", unstuffed->string[i] & 0xff );
    }
    if ( unstuffed->dfile_size == 0 ) { 
	elog_fprintf(file, "    *trailer packet*\n" ) ; 
    } else { 
	elog_fprintf(file, "    ndata    = %8d\n", unstuffed->dfile_size ) ; 
    }
    if ( mode == PKT_PEEK ) { 
	hexdump (file, packet, MIN(48, nbytes)) ; 
    }
}

static void
show_db ( FILE *file, Dbptr db, int mode )  
{
    char *table_name, *field_name ; 
    char field[STRSZ], null[STRSZ] ; 
    long nfields, line, size ; 
    Dbptr db2 ;
    dbquery ( db, dbTABLE_NAME, &table_name ) ;
    elog_fprintf ( file, "    Table %s\n", table_name ) ;
    switch (mode) { 
      case PKT_PEEK:
	dbquery ( db, dbFIELD_COUNT, &nfields ) ; 
	line = 0 ; 
	for ( db.field = 0 ; db.field < nfields ; db.field++ ) {
	    dbget ( db, field ) ; 
	    db2 = db ; db2.record = dbNULL ;
	    dbget ( db2, null ) ;
	    if ( strcmp ( field, null ) != 0 ) { 
		dbquery ( db, dbFIELD_NAME, &field_name ) ; 
		dbquery ( db, dbFIELD_SIZE, &size ) ; 
		if ( line+size > 100 ) { 
		    break ;
		}
		if ( strlen(field_name) < size ) { 
		    elog_fprintf(file, "%-*s ", (int) size, field_name ) ; 
		} else { 
		    strncpy(null, field_name, size) ;
		    null[size] = 0 ; 
		    elog_fprintf(file, "%-*s ", (int) size, null ) ; 
		}
		line += size ;
	    }
	}
	nfields = db.field+1 ;

	for ( db.field = 0 ; db.field < nfields ; db.field++ ) {
	    dbget ( db, field ) ; 
	    db2 = db ; db2.record = dbNULL ;
	    dbget ( db2, null ) ;
	    if ( strcmp ( field, null ) != 0 ) { 
		dbquery ( db, dbFIELD_NAME, &field_name ) ; 
		elog_fprintf ( file, "    %s = '%s'\n", field_name, field ) ; 
	    }
	}
	break; 

      default:
      case PKT_UNSTUFF:
	dbquery ( db, dbFIELD_COUNT, &nfields ) ; 
	for ( db.field = 0 ; db.field < nfields ; db.field++ ) {
	    dbget ( db, field ) ; 
	    db2 = db ; db2.record = dbNULL ;
	    dbget ( db2, null ) ;
	    if ( strcmp ( field, null ) != 0 ) { 
		dbquery ( db, dbFIELD_NAME, &field_name ) ; 
		elog_fprintf ( file, "    %s = '%s'\n", field_name, field ) ; 
	    }
	}
	break; 
    }
}

static void
show_pf (FILE *file, Pf *pf, int mode ) 
{
    Tbl *list ; 
#define NCOLUMNS  4
    switch (mode) { 
      case PKT_PEEK:
	list = pfkeys(pf) ; 
	for ( int i = 0 ; i<maxtbl(list) ; i++ ) { 
	    char *akey = gettbl(list, i) ; 
	    if ( i>0 && i % NCOLUMNS == 0 ) { 
		fprintf(file, "\n" ) ; 
	    }
	    if ( i==0 || i % NCOLUMNS == 0 ) { 
		fprintf(file, "    " ) ; 
	    }
	    fprintf(file, " %-15s", akey) ; 
	}
	fprintf(file, "\n" ) ;
	freetbl(list,0) ; 
	break ;
      case PKT_NOSAMPLES:
	break ;
      default :
      case PKT_UNSTUFF:
	pfout(file, pf) ;
	break;
    }
}


static void
showPktSpecific(FILE *file, Packet *pkt) 
{
    if ( strcmp(pkt->parts.src_suffix, "CD1") && strcmp(pkt->parts.src_suffix, "CD1S") ) return;
	if ( pkt->string != 0 ) { 
	    elog_fprintf(file, "weather: %s\n", okstr(pkt->string) ) ;
	}
}

static char *setStatus[] = {
    "signature at digitizer",		/*  0 */
    "**error**",			/*  1 */
    "**error**",			/*  2 */
    "**error**",			/*  3 */
    "**error**",			/*  4 */
    "**error**",			/*  5 */
    "**error**",			/*  6 */
    "**error**",			/*  7 */
    "**error**",			/*  8 */
    "**error**",			/*  9 */
    0,					/* 10 */
    0,					/* 11 */
    0,					/* 12 */
    0,					/* 13 */
    0,					/* 14 */
    0,					/* 15 */
    "digitizer calibrator loopback on",	/* 16 */
    "digitizer analog input shorted",	/* 17 */
    "Intrasite communications buffer overflow",		/* 18 */
    "GPS antenna or receiver off",	/* 19 */
    "clock differential too large", 	/* 20 */
    "equipment moved", 			/* 21 */
    "authentication unit opened", 	/* 22 */
    "vault door open",			/* 23 */
    "**error**",			/* 24 */
    "**error**",			/* 25 */
    0,					/* 26 */
    0,					/* 27 */
    "calibration signal", 		/* 28 */
    "data clipped by digitizer", 			/* 29 */
    "zeroed data from digitizer", 		/* 30 */
    "dead channel from digitizer",		/* 31 */
} ;

static void
show_CD1S_status ( FILE *file, char *status ) 
{
    int stat, i, bit, voltage ;
    char *statusp ; 
    /* memcpy ( &stat, status, 4 ) ; */
    statusp = status ; 
    umi2hi ( &statusp, (unsigned int *)&stat, 1 ) ; 
    for ( i=31 ; i>=0 ; i-- ) { 
	bit = 1 << i ; 
	if ( bit & stat ) { 
	    if ( setStatus[i] != 0 ) { 
		elog_fprintf ( file, "bit #%2d: %s\n", i, setStatus[i] ) ;
	    }
	} 
    }
    if ( stat & (1 <<28) ) { 
	switch ( (stat>>26) & 3 ) { 
	    case 0:
		elog_fprintf ( file, "pulse calibration\n" ) ;
		break ;

	    case 1:
		elog_fprintf ( file, "step calibration\n" ) ;
		break ;
	  
	    case 2:
		elog_fprintf ( file, "sinewave calibration\n" ) ;
		break ;

	    case 3:
		elog_fprintf ( file, "random binary calibration\n" ) ;
		break ;
	}
    }
    voltage = (stat>>10) & 0x3f ;
    elog_fprintf ( file, "Digitizer Voltage: 0x%x\n", voltage ) ;
}

static void
showPktChannelSpecific(FILE *file, PktChannel *achan, Packet *pkt) 
{
    if ( ! strcmp(pkt->parts.src_suffix, "CD1") || ! strcmp(pkt->parts.src_suffix, "CD1S") ) {
		switch ( achan->iuser1 ) { 
	    	case 1:
			elog_fprintf ( file, "compressed, " ) ;
			break ;
	
	    	case 0:
			elog_fprintf ( file, "not compressed, " ) ;
			break ;
	
	    	default:
			elog_fprintf ( file, "invalid compression flag: %d, ", achan->iuser1);
			break ;
		}

		switch ( achan->iuser2 ) { 
	    	case 1:
			elog_fprintf ( file, "signed:\n" ) ;
			hexdump ( file, achan->cuser2, 40 ) ;
			break ;
	
	    	case 0:
			elog_fprintf ( file, "not signed\n" ) ;
			break ;
	
	    	default:
			elog_fprintf ( file, "invalid signed flag: %d\n", achan->iuser2);
			break ;
		}

		show_CD1S_status ( file, achan->cuser1 ) ; 
	}
}


static void
fchanshow ( FILE *file, int chan, PktChannel *achan, int samples, Packet *pkt ) 
{
    int i ;
    char *s, *s2 ;
    float *data ;
    double t ; 
    char *fmt1 = "%a %Y-%j %b %d %k:%M:%S.%u", 
    	 *fmt2 = "%k:%M:%S.%u" ;

    t = dround(achan->time*1.e5)*1e-5 ;
    elog_fprintf ( file, "    %2d : %-8s %-8s %-8s %-8s %6.3f/s calib=%12.8g calper=%6.3f segtype=%s %d samps %s - %s",
	chan, 
	achan->net, 
	achan->sta, 
	achan->chan, 
	achan->loc, 
	achan->samprate, 
	achan->calib,
	achan->calper,
	achan->segtype,
	achan->nsamp, 
	s=epoch2str(t, fmt1), 
	s2=epoch2str(SAMP2TIME(t, achan->samprate, achan->nsamp),fmt2)
	) ;
    free(s) ;
    free(s2) ;
    if ( FullPrecision ) { 
	elog_fprintf(file, " (%17.7f)\n", achan->time ) ; 
    } else { 
	elog_fprintf(file, "\n" ) ; 
    }
    showPktChannelSpecific(file, achan, pkt) ;
    if (samples != 0) {
	int n ; 
	if ( samples < 0 ) { 
	    n = achan->nsamp ; 
	} else { 
	    n = MIN(samples, achan->nsamp) ; 
	}
	data = (float *) achan->data ;
	elog_fprintf ( file, "\t" ) ;
	for (i=0 ; i<n ; i++ ) {
	    if ( i>0 && (i % 10) == 0 ) elog_fprintf ( file, "\n\t" ) ; 
	    elog_fprintf ( file, " %10g", data[i] ) ; 
	}
	elog_fprintf ( file, "\n" ) ; 
    }
}

static void
chanshow ( FILE *file, int chan, PktChannel *achan, int samples, Packet *pkt ) 
{
    int i ;
    char *s, *s2 ;
    int *data ;
    double t ; 
    char *fmt1 = "%a %Y-%j %b %d %k:%M:%S.%u", 
    	 *fmt2 = "%k:%M:%S.%u" ;

    t = dround(achan->time*1.e5)*1e-5 ;
    elog_fprintf ( file, "    %2d : %-8s %-8s %-8s %-8s %6.3f/s calib=%12.8g calper=%6.3f segtype=%s %d samps %s - %s",
	chan, 
	achan->net, 
	achan->sta, 
	achan->chan, 
	achan->loc, 
	achan->samprate, 
	achan->calib,
	achan->calper,
	achan->segtype,
	achan->nsamp, 
	s=epoch2str(t, fmt1), 
	s2=epoch2str(SAMP2TIME(t, achan->samprate, achan->nsamp),fmt2)
	) ;
    free(s) ;
    free(s2) ;
    if ( FullPrecision ) { 
	elog_fprintf(file, " (%17.7f)\n", achan->time ) ; 
    } else { 
	elog_fprintf(file, "\n" ) ; 
    }
    showPktChannelSpecific(file, achan, pkt) ;
    if (samples != 0) {
	int n ; 
	if ( samples < 0 ) { 
	    n = achan->nsamp ; 
	} else { 
	    n = MIN(samples, achan->nsamp) ; 
	}
	data = achan->data ;
	elog_fprintf ( file, "\t" ) ;
	for (i=0 ; i<n ; i++ ) {
	    if ( i>0 && (i % 10) == 0 ) elog_fprintf ( file, "\n\t" ) ; 
	    elog_fprintf ( file, " %8d", data[i] ) ; 
	}
	elog_fprintf ( file, "\n" ) ; 
    }
}



void
showPktMatch ( int pktid, char *srcname, double pkttime, char *packet, int nbytes, 
	FILE *file, char *match, int mode )  
{
    Packet *unstuffed=0 ;
    int i, samples = -1 ;
    PktChannel *achan ;
    char *s, *msg ;
    Hook *hook = 0 ; 
    int           (*amethod)() ;
    int	result ; 
    long	start, nchars ; 
    char stachan[64] ; 
    /* char *fmt1 = "%m/%d/%Y (%j) %k:%M:%S.%u" ;  */

    msg = elogstr ( ELOG_DEBUGGING, 0, "#%d '%s': %s : %d bytes%s\n", 
	    pktid, srcname, 
	    s = strydtime(pkttime),
	    /* s=epoch2str(pkttime,fmt1), */
	    nbytes, 
	    is_stash(srcname,packet,nbytes) ? " *stash*" : "" ) ;
    free(s) ;
    fputs(msg, file) ;
    free(msg) ;

    switch (mode) {
	case PKT_DUMP:
	    elog_fprintf ( file, "%s %s %s %d %.7f\n", 
		    "#orb", "x", srcname, nbytes, pkttime ) ;
	    hexdump ( file, packet, nbytes ) ; 
	    break ;

	default:
	case PKT_NOSAMPLES:
	case PKT_UNSTUFF:
	case PKT_PEEK:
	    result = unstuffPkt ( srcname, pkttime, packet, nbytes, &unstuffed ) ;
	    switch ( result ) { 
		case Pkt_wf :
		    switch (mode) { 
			case PKT_PEEK:
			    samples = 10 ; 
			    break; 
			case PKT_NOSAMPLES:
			    samples = 0 ; 
			    break; 
			case PKT_UNSTUFF:
			    samples = -1 ; 
			    break ;
		    }
		    showPktSpecific(file, unstuffed) ;
		    for ( i=0 ; i<unstuffed->nchannels ; i++ ) {
				achan = (PktChannel *) gettbl(unstuffed->channels, i) ;
				sprintf(stachan, "%s_%s", achan->sta, achan->chan ) ; 
				if (match == 0 || *match == 0 || strmatches(stachan, match, &hook)) { 
			    	chanshow ( file, i, achan, samples, unstuffed ) ; 
				}
		    }
			if (!strcmp(unstuffed->parts.src_suffix, "CD11S")) {
				cd11s_show (file, (unsigned char *)(packet+20), "       ");
			} else if (!strcmp(unstuffed->parts.src_suffix, "CD11")) {
			    unsigned char *cd11_frame;
			    int cd11_frame_nbytes;
			    if (cd11_unstuff_frame (packet, nbytes, &cd11_frame, &cd11_frame_nbytes) < 0) {
			    	complain (0, "cd11_unstuff_frame() error.\n");
			    } else {
			    	cd11_show (file, cd11_frame, "       ");
			    }
			}
		    break ;

		case Pkt_float :
		    switch (mode) { 
			case PKT_PEEK:
			    samples = 10 ; 
			    break; 
			case PKT_NOSAMPLES:
			    samples = 0 ; 
			    break; 
			case PKT_UNSTUFF:
			    samples = -1 ; 
			    break ;
		    }
		    showPktSpecific(file, unstuffed) ;
		    for ( i=0 ; i<unstuffed->nchannels ; i++ ) {
			achan = (PktChannel *) gettbl(unstuffed->channels, i) ;
			sprintf(stachan, "%s_%s", achan->sta, achan->chan ) ; 
			if (match == 0 || *match == 0 || strmatches(stachan, match, &hook)) { 
			    fchanshow ( file, i, achan, samples, unstuffed ) ; 
			}
		    }
		    break ;

		case Pkt_st:			       /* status packet */
		    elog_fprintf ( file, "    status packet\n" ) ;
		    hexdump ( file, packet, nbytes ) ; 
		    break;

		case Pkt_stash:			       /* stash packet */
		    elog_fprintf ( file, "    stash packet\n" ) ;
		    show_stash ( file, unstuffed, mode ) ; 
		    break;
	 
		 case Pkt_db:
		    elog_fprintf(file, "    database row\n" ) ;
		    show_db ( file, unstuffed->db, mode ) ; 
		    if ( unstuffed->dfile ) { 
			elog_fprintf (file, "dfile:\n" ) ;
			fputs (unstuffed->dfile, file ) ; 
			elog_fprintf (file, "\n" ) ;
		    }
		    break;

		case Pkt_pf:
		    elog_fprintf ( file, "    parameter file data packet\n" ) ;
		    show_pf(file, unstuffed->pf, mode ) ; 
		    break;

		case Pkt_cn:
		    elog_fprintf ( file, "    control message packet\n" ) ;
		    pfout(file, unstuffed->pf) ;
		    break;

		case Pkt_msg:
		    elog_fprintf ( file, "    control message packet\n" ) ;
		    elog_fprintf(file, "Envelope\n") ; 
		    pfout(file, unstuffed->pf) ;
		    elog_fprintf(file, "%s\n", unstuffed->string) ; 
		    break;

		case Pkt_rw:
		    elog_fprintf ( file, "    raw data packet\n" ) ;
		    if ( mode == PKT_PEEK ) { 
			hexdump ( file, packet, MIN(48, nbytes) ) ; 
		    } else { 
			hexdump ( file, packet, nbytes ) ; 
		    }
		    break;

		case Pkt_ch:
		    elog_fprintf ( file, "    string packet\n" ) ; 
		    if ( unstuffed->string != 0 ) { 
			if ( mode == PKT_PEEK ) { 
			    char *s, tmp[256] ; 
			    memcpy(tmp, unstuffed->string, sizeof(tmp)) ;
			    strcpy(tmp+sizeof(tmp)-6, "\n\t...\n") ;
			    s = strchr(tmp, '\n') ; 
			    if ( s != 0 ) { 
				s = strchr(s+1, '\n') ;
			    }
			    if ( s != 0 ) {
				s = strchr(s+1, '\n') ; 
			    }
			    if ( s != 0 ) { 
				*s = 0 ; 
			    }
			    elog_fprintf ( file, "    '%s'\n", tmp ) ;
			} else { 
			    elog_fprintf ( file, "    '%s'\n", unstuffed->string ) ;
			}
		    }
		    break;

		case Pkt_fl:
		    show_fl (pktid, srcname, pkttime, packet, nbytes, unstuffed, file, mode ) ;
		    break;

		default: 
		    amethod = (int (*)()) usermethod("showPkt_User") ;
		    if ( amethod != 0 ) {
			amethod (pktid, srcname, pkttime, packet, nbytes, unstuffed, file, mode) ;
		    } else {
			int n ;
			n = elog_mark() ; 
			if ( n > 0 ) { 
			    complain ( 0, "*** couldn't unstuff packet ***\n" ) ; 
			} else { 
			    elog_fprintf ( file, "*** couldn't unstuff packet ***\n" ) ; 
			}
			elog_fprintf ( file, "%s %s %s %d %.7f\n", 
				"#orb", "x", srcname, nbytes, pkttime ) ;
			if ( mode == PKT_PEEK ) { 
			    hexdump (file, packet, MIN(48, nbytes)) ; 
			}
			if ( strcontains(srcname, "/pf/", 0, &start, &nchars ) )  { 
			    complain ( 0, "*** couldn't unstuff packet ***\n" ) ; 
			    fprintf(stderr, "uncompiled pf string:\n%s\n\n", okstr(unstuffed->string)) ; 
			}
		    }
		    break ;
	    }
	    freePkt(unstuffed) ;
	    break ;

	case PKT_TERSE:
	    break ;
    }
    fflush(file) ;
}

void
showPkt ( int pktid, char *srcname, double pkttime, char *packet, int nbytes, 
	FILE *file, int mode )  
{
    showPktMatch ( pktid, srcname, pkttime, packet, nbytes, file, 0, mode ) ; 
}
    

char *
strshowPkt ( int pktid, char *srcname, double pkttime, char *packet, int nbytes ) 
{
    char msg[96], *s ; 
    snprintf ( msg, sizeof(msg), "#%d '%s': %s : %d bytes%s", 
	    pktid, srcname, 
	    s = strydtime(pkttime),
	    nbytes, 
	    is_stash(srcname,packet,nbytes) ? " *stash*" : "" ) ;
   free(s) ; 
   return strdup(msg) ; 
}

