//   Copyright (c) 2016 Boulder Real Time Technologies, Inc.           
//                                                                     
//   This software module is wholly owned by Boulder Real Time         
//   Technologies, Inc. Any use of this software module without        
//   express written permission from Boulder Real Time Technologies,   
//   Inc. is prohibited.                                               

#ifndef CD11_FRAME_H
#define CD11_FRAME_H

#include "oobns.h"

class CD11FrameHeader;
class CD11FramePayload;
class CD11FrameTrailer;
class GapSet;

class CD11Frame
{
	public: 	

						CD11Frame( void );
						CD11Frame( CD11FrameType type, std::string sta, std::string statype, int rqip, unsigned short rqport );	// ConnRequest
						CD11Frame( CD11FrameType type, int responderip, unsigned short responderport );				// ConnResponse
						CD11Frame( CD11FrameType type, int optionType, int size, std::string request );				// OptionRequest
						CD11Frame( CD11FrameType type, CD11OptionRequestPayload* orqPayload );					// OptionResponse
						CD11Frame( CD11FrameType type, std::shared_ptr<OooOrbPacket> op );					// Data
						CD11Frame( CD11FrameType type, std::string message );							// Alert
						CD11Frame( CD11FrameType type, GapSet* gapSet );							// AckNack
						CD11Frame( OOBns* oobns );
						CD11Frame( std::istream& is );
						~CD11Frame( void );

		void				resizePayload( int payloadSize );	
		void				resizeTrailer( int trailerSize );	

		int 				frameSize( void );	
		std::string 			frameTypeString( bool quoted = false );
		std::string 			frameSetName( void );

		unsigned char*			frameBufPtr( void );
		unsigned char*			headerBufPtr( void );
		unsigned char*			payloadBufPtr( void );
		unsigned char*			trailerBufPtr( void );

		CD11FrameHeader*		header( void );
		CD11FramePayload*		payload( void );			
		CD11FrameTrailer*		trailer( void );

		CD11ConnRequestPayload*		crqPayload( void );			
		CD11ConnResponsePayload*	crsPayload( void );			
		CD11OptionRequestPayload*	orqPayload( void );			
		CD11OptionResponsePayload*	orsPayload( void );			
		CD11DataPayload*		dataPayload( void );			
		CD11AlertPayload*		alertPayload( void );			
		CD11AckNackPayload*		acknackPayload( void );			

		std::string 			toString( std::string headerMessage = "" );

		void 				print( std::ostream& os );
		void 				write( std::ostream& os );

		// friend std::ostream&		operator<<( std::ostream& os, std::shared_ptr<CD11Frame>& f );

		virtual void			marshal( OOBns* oobns );

		static int 			pad4( int v );
		static std::string		type2string( CD11FrameType type, bool quoted = false );

	private:
		
		void				resize( void );	

		unsigned char*			m_nbo_frameBuf;
		unsigned char*			m_nbo_headerBufPtr;
		unsigned char*			m_nbo_payloadBufPtr;
		unsigned char*			m_nbo_trailerBufPtr;

		int 				m_frameBufSize;
		int				m_headerBufSize;
		int				m_payloadBufSize;
		int				m_trailerBufSize;

		CD11FrameHeader*		m_header;
		CD11FramePayload*		m_payload;
		CD11FrameTrailer*		m_trailer;
};

#endif

