//   Copyright (c) 2016 Boulder Real Time Technologies, Inc.           
//                                                                     
//   This software module is wholly owned by Boulder Real Time         
//   Technologies, Inc. Any use of this software module without        
//   express written permission from Boulder Real Time Technologies,   
//   Inc. is prohibited.                                               

#ifndef CD11_FRAME_HEADER_H
#define CD11_FRAME_HEADER_H

#include <iostream>

#include "oobns.h"

#define CD11_FRAME_CREATOR_SIZE 8
#define CD11_FRAME_DESTINATION_SIZE 8
#define CD11_FRAME_HEADER_SIZE ( 3 * sizeof( int ) + sizeof( int64_t ) + CD11_FRAME_CREATOR_SIZE + CD11_FRAME_DESTINATION_SIZE )

class CD11Frame;

enum CD11FrameType : int;

class CD11FrameHeader
{
	public: 	

					CD11FrameHeader( CD11Frame* frame );
					~CD11FrameHeader( void );

		void			resetPointers( void );

		void			setFromPayload( CD11FrameType type );

		int			headerSize( void );
		
		CD11FrameType		frameType( void );
		int 			trailerOffset( void );
		std::string		frameCreator( void );
		std::string		frameDestination( void );
		int64_t 		sequenceNumber( void );
		int			seriesNumber( void );

		void 			frameType( CD11FrameType v );
		void 			trailerOffset( int v );
		void 			frameCreator( std::string v );
		void			frameDestination( std::string v );
		void 			sequenceNumber( int64_t v );
		void			seriesNumber( int v );

		void			unmarshal( OOBns* oobns );
		void 			read( std::istream& is );

		std::string 		toString( void );
		
		void			print( std::ostream& os );

	private:

		int64_t 		epochToSequence( double epoch );

		CD11Frame*		m_frame;
		int 			m_headerSize;

		int*			m_nbo_frameType;
		int*			m_nbo_trailerOffset;
		char*			m_nbo_frameCreator;
		char*			m_nbo_frameDestination;
		int64_t*		m_nbo_sequenceNumber;
		int*			m_nbo_seriesNumber;
};

#endif


