//   Copyright (c) 2016 Boulder Real Time Technologies, Inc.           
//                                                                     
//   This software module is wholly owned by Boulder Real Time         
//   Technologies, Inc. Any use of this software module without        
//   express written permission from Boulder Real Time Technologies,   
//   Inc. is prohibited.                                               

#ifndef CD11_FRAME_TRAILER_H
#define CD11_FRAME_TRAILER_H

#include <iostream>

#include "oobns.h"
#include "versign.h"

#define CD11_FRAME_TRAILER_MIN_SIZE ( 2 * sizeof( int ) + sizeof( int64_t ) )

class CD11Frame;

class CD11FrameTrailer
{
	public: 	
					CD11FrameTrailer( void );
					CD11FrameTrailer( CD11Frame* frame );
					~CD11FrameTrailer( void );

		void			resetPointers( void );

		int			trailerSize( void );

		void			sign( void );
		void			authenticate( void );
		void			addCRC( void );
		void			verifyCRC( void );

		uint32_t		authKeyIdentifier( void );
		int			authSize( void );
		unsigned char* 		authValue( void );
		uint64_t 		commCRC( void );

		void			authKeyIdentifier( uint32_t v );
		void			authSize( int v );
		void			authValue( unsigned char* v, int n );
		void			commCRC( uint64_t v );

		void			unmarshal( OOBns* oobns );
		void 			read( std::istream& is );

		std::string 		toString( void );
		
		void			print( std::ostream& os );

	private:

		static void 		generateChecksumTable( void );
		static uint64_t 	calculateCRC( const unsigned char* buffer, const uint32_t buflen );

		static uint64_t		s_generatorPolynomial;
		static uint64_t 	s_checksumTable[256];
		static bool 		s_checksumInitialized;

		CD11Frame*		m_frame;
		int 			m_trailerSize;

		int*			m_nbo_authKeyIdentifier;
		int*			m_nbo_authSize;
		unsigned char*		m_nbo_authValue;	
		int64_t*		m_nbo_commCRC;
};

#endif


