//   Copyright (c) 2016 Boulder Real Time Technologies, Inc.           
//                                                                     
//   This software module is wholly owned by Boulder Real Time         
//   Technologies, Inc. Any use of this software module without        
//   express written permission from Boulder Real Time Technologies,   
//   Inc. is prohibited.                                               

#ifndef CD11_SOCKET_H
#define CD11_SOCKET_H

#include <stdio.h>
#include <errno.h>
#include <resolv.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>

class CD11Socket
{
	public:
							CD11Socket( void );
							~CD11Socket( void );
	
		void					closeSocket( void );			

		int 					getLocalIP( void );	
		unsigned short 				getLocalPort( void );	
		std::string 				getLocalEndpoint( void );	
		std::string 				getPeerEndpoint( void );	

	protected:

		void 					queryLocalAddress( void );
		void 					setPeerFromAddr( void );
		
		int 					m_fd;
		struct sockaddr_in 			m_localAddr;	
		std::string 				m_localHostname;
		unsigned short				m_localPort;
		struct sockaddr_in 			m_peerAddr;	
		std::string 				m_peerHostname;
		unsigned short				m_peerPort;
		OOBns* 					m_oobnsIn;
		OOBns* 					m_oobnsOut;
};

class CD11ConnectedSocket : public CD11Socket
{
	public:
							CD11ConnectedSocket( int fd, struct sockaddr_in addr, double timeout );
							CD11ConnectedSocket( std::string hostname, unsigned short port, double timeout );
							CD11ConnectedSocket( int ip, unsigned short port, double timeout );
							~CD11ConnectedSocket( void );

		void 					sendCD11Frame( std::shared_ptr<CD11Frame> );
		std::shared_ptr<CD11Frame> 		recvCD11Frame( void );
		Tbl* 					getFrameSets( void );

	protected: 
	
		void 					init( int ip, double timeout );

		Table<std::string>			m_receivingFrameSets;

};

class CD11ListenerSocket : public CD11Socket
{
	public:
							CD11ListenerSocket( int ip, unsigned short port, int backlog );
							~CD11ListenerSocket( void );	

		CD11ConnectedSocket*			accept_timeout( double timeout );

	protected:

		int 					m_backlog;
};

#endif